from setuptools import setup


setup(
    name='hoser',
    version='0.0.4',
    description='Build data pipelines fast',
    author='Mark Asp',
    url='https://github.com/hoser-io/hoser-py',
    packages=['hoser', 'hoser.stdlib'],
)